"""
dxf2laser V0.4alpha
Calculation of gcode for pcb-exposure with a UV-laser on a CNC-machine.
Datasource: dxf-export from KiCAD 4.0.2 stable
Datatarget: G-Code for LinuxCNC
Tested in python 2.7.6
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <FlorenzW> wrote this file. As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return. F. Wimmer
 * ----------------------------------------------------------------------------
"""

# -*- coding: utf-8 -*-
from Tkinter import *
import math
from copy import deepcopy
import fpformat
from random import randint

xMin = 0
yMin = 0
xMax = 0
yMax = 0
scaling = 0

# Size of the grafical output
xCanvas = 1024
yCanvas = 768
canvas = Canvas(width=xCanvas, height=yCanvas, bg='white')  
canvas.pack(expand=YES, fill=BOTH)                  


def export_gcode(geometriy_outlines, filling_pattern):
    print("Exporting GCode")
    DateiOutput = open(filename_output,"w")

    DateiOutput.write("G21")
    DateiOutput.write("\n")
    DateiOutput.write("G90")
    DateiOutput.write("\n")
    DateiOutput.write("G94")
    DateiOutput.write("\n")
    DateiOutput.write("G64 P0.001")
    DateiOutput.write("\n")
    DateiOutput.write("M5")
    DateiOutput.write("\n")
    DateiOutput.write("F1000")
    DateiOutput.write("\n")
    DateiOutput.write("G0 X0.0 Y0.0 Z0.0")
    DateiOutput.write("\n")

    i = 0
    while (i < len(geometriy_outlines)):
        DateiOutput.write("G0 X")
        DateiOutput.write(fpformat.fix(geometriy_outlines[i][0][0], 3))
        DateiOutput.write(" Y")
        DateiOutput.write(fpformat.fix(geometriy_outlines[i][0][1], 3))
        DateiOutput.write("\n")
        DateiOutput.write("M3")
        DateiOutput.write("\n")

        j = 1
        while (j < len(geometriy_outlines[i])):
            DateiOutput.write("G1 X")
            DateiOutput.write(fpformat.fix(geometriy_outlines[i][j][0], 3))
            DateiOutput.write(" Y")
            DateiOutput.write(fpformat.fix(geometriy_outlines[i][j][1], 3))
            DateiOutput.write("\n")
            j = j + 1

        DateiOutput.write("G1 X")
        DateiOutput.write(fpformat.fix(geometriy_outlines[i][0][0], 3))
        DateiOutput.write(" Y")
        DateiOutput.write(fpformat.fix(geometriy_outlines[i][0][1], 3))
        DateiOutput.write("\n")
        DateiOutput.write("M5")
        DateiOutput.write("\n")

        i = i + 1    
            

    i = 0
    while i < len(filling_pattern):
        DateiOutput.write("G0 X")
        DateiOutput.write(fpformat.fix(filling_pattern[i][0][0], 3))
        DateiOutput.write(" Y")
        DateiOutput.write(fpformat.fix(filling_pattern[i][0][1], 3))
        DateiOutput.write("\n")
        DateiOutput.write("M3")
        DateiOutput.write("\n")
        DateiOutput.write("G1 X")
        DateiOutput.write(fpformat.fix(filling_pattern[i][1][0], 3))
        DateiOutput.write(" Y")
        DateiOutput.write(fpformat.fix(filling_pattern[i][1][1], 3))
        DateiOutput.write("\n")
        DateiOutput.write("M5")
        DateiOutput.write("\n")

        i = i + 1


    DateiOutput.write("\n")
    DateiOutput.write("M2")
    DateiOutput.write("\n")
    DateiOutput.close()



def draw_fillingpattern_on_canvas(fillingpattern, color):
    print("Drawing fillingpattern on canvas")
    i = 0
    while (i < len(fillingpattern)):
        canvas.create_line(fillingpattern[i][0][0],fillingpattern[i][0][1],fillingpattern[i][1][0],fillingpattern[i][1][1], fill = color)
        i = i + 1    


def map_fillingpattern_on_canvas(fillingpattern):
    pattern = deepcopy(fillingpattern)
    i = 0
    while (i < len(pattern)):
        pattern[i][0] = map_point_on_canvas(pattern[i][0])
        pattern[i][1] = map_point_on_canvas(pattern[i][1])
        i = i + 1
    return pattern


def eliminate_vertical_lines(all_lines):
    i = 0
    lines = deepcopy(all_lines)
    while i < len(lines):
        if lines[i][0][0] == lines[i][1][0]:
            del(lines[i])
            i = -1
        i = i + 1
    return lines
        

def calculate_filling_pattern(all_lines):
    print("Calculating fillingpattern - this takes a while...")
    global xMin
    global xMax
    global yMin
    global yMax
    global distance_fillinglines
    x = xMin
    fillingpattern = []
    filling_orientation = True   #is switched every line to turn the orientation

    while x < xMax:
        i = 0
        intersections = []
        while i < len(all_lines):

            pointK = []
            pointK.append(x)
            pointK.append(0)
            pointL = []
            pointL.append(x)
            pointL.append(1)
            pointM = []
            pointM = all_lines[i][0]
            pointN = []
            pointN = all_lines[i][1]
            
            if ((pointM[0] < x) and (pointN[0] > x)) or ((pointM[0] > x) and (pointN[0] < x)):
                pointS = calculate_intersection(pointK, pointL, pointM, pointN)
                intersections.append(pointS[1])

            i = i + 1

        if filling_orientation:
            intersections.sort(reverse = False)

            line = []
            point = []
            point.append(x)
            point.append(yMin)
            line.append(point)
            j = 0
            while j < len(intersections):
                point = []
                point.append(x)
                point.append(intersections[j])
                line.append(point)
                fillingpattern.append(line)
                line = []
                point = []
                point.append(x)
                point.append(intersections[j+1])
                line.append(point)

                j = j + 2

            point = []
            point.append(x)
            point.append(yMax)
            line.append(point)
            fillingpattern.append(line)

        else:
            intersections.sort(reverse = True)

            line = []
            point = []
            point.append(x)
            point.append(yMax)
            line.append(point)
            j = 0
            while j < len(intersections):
                point = []
                point.append(x)
                point.append(intersections[j])
                line.append(point)
                fillingpattern.append(line)
                line = []
                point = []
                point.append(x)
                point.append(intersections[j+1])
                line.append(point)

                j = j + 2

            point = []
            point.append(x)
            point.append(yMin)
            line.append(point)
            fillingpattern.append(line)
            
        x = x + distance_fillinglines

        filling_orientation = not filling_orientation  #Richtungswechsel bei jeder Zeile -> Wegoptimierung

    return fillingpattern
            

def explode_to_lines(all_polylines):
    i = 0
    lines = []
    while i < len(all_polylines):
        j = 0
        while j < len(all_polylines[i]):
            line = []
            line.append(all_polylines[i][j])
            if j < len(all_polylines[i])-1:
                line.append(all_polylines[i][j+1])
            else:
                line.append(all_polylines[i][0])
            lines.append(line)
            j = j + 1
        i = i + 1
    return lines


def calculate_positioning_length_outlines(all_polylines):
    i = 0
    length = 0
    while i < len(all_polylines)-1:
        length = length + line_length([all_polylines[i][0],all_polylines[i+1][0]])
        i = i + 1
    return length


def optimizing_outline_arrangement(all_polylines):
    global optimizing_cycles
    print("Optimizing outline arrangement, this takes a while...")
    print("Length old: "),
    print(calculate_positioning_length_outlines(all_polylines))

    t = 0
    while t < optimizing_cycles:

        i = 0
        while i < 1000:

            length_old = calculate_positioning_length_outlines(all_polylines)
            random_posA = randint(0,len(all_polylines)-1)
            random_posB = randint(0,len(all_polylines)-1)
            all_polylines[random_posB], all_polylines[random_posA] = all_polylines[random_posA], all_polylines[random_posB]
            if length_old <= calculate_positioning_length_outlines(all_polylines): 
                all_polylines[random_posB], all_polylines[random_posA] = all_polylines[random_posA], all_polylines[random_posB]

            i = i + 1

        t = t + 1
        
    print("Length new: "),
    print(calculate_positioning_length_outlines(all_polylines))


def calculate_outline(all_polylines, laser_diameter):
    temp_all_polylines = deepcopy(all_polylines)
    i = 0
    while i < len(all_polylines):
        j = 0
        polyline_outline = []
        while j < len(all_polylines[i]):
            
            pointA = all_polylines[i][j]

            if (j == len(all_polylines[i])-1):
                pointB = all_polylines[i][0]
            else:
                pointB = all_polylines[i][j+1]

            if (j < len(all_polylines[i])-2):
                pointC = all_polylines[i][j+2]
            if (j == len(all_polylines[i])-2):
                pointC = all_polylines[i][0]
            if (j == len(all_polylines[i])-1):
                pointC = all_polylines[i][1]

            vectorA = []
            vectorA.append(pointB[1]-pointA[1])
            vectorA.append(pointB[0]-pointA[0])
            lengthA = line_length([[0,0], vectorA])
                  
            vectorA[0] = vectorA[0] / lengthA * (laser_diameter / 2)
            vectorA[1] = vectorA[1] / lengthA * (laser_diameter / 2)
        
            vectorB = []
            vectorB.append(pointC[1]-pointB[1])
            vectorB.append(pointC[0]-pointB[0])
            lengthB = line_length([[0,0], vectorB])

            vectorB[0] = vectorB[0] / lengthB * (laser_diameter / 2)
            vectorB[1] = vectorB[1] / lengthB * (laser_diameter / 2)
        

            pointK=[]
            pointK.append(pointA[0]-vectorA[0])
            pointK.append(pointA[1]+vectorA[1])

            pointL=[]
            pointL.append(pointB[0]-vectorA[0])
            pointL.append(pointB[1]+vectorA[1])

            pointM=[]
            pointM.append(pointB[0]-vectorB[0])
            pointM.append(pointB[1]+vectorB[1])

            pointN=[]
            pointN.append(pointC[0]-vectorB[0])
            pointN.append(pointC[1]+vectorB[1])

            pointS = calculate_intersection(pointK, pointL, pointM, pointN)

            temp_all_polylines[i][j] = (pointS)

            j = j + 1

        i = i + 1

    return temp_all_polylines


def calculate_intersection(pointK, pointL, pointM, pointN):
    pointS=[]
    pointS.append(((pointN[0]-pointM[0])*(pointL[0]*pointK[1]-pointK[0]*pointL[1])-(pointL[0]-pointK[0])*(pointN[0]*pointM[1]-pointM[0]*pointN[1]))/((pointN[1]-pointM[1])*(pointL[0]-pointK[0])-(pointL[1]-pointK[1])*(pointN[0]-pointM[0])))
    pointS.append(((pointK[1]-pointL[1])*(pointN[0]*pointM[1]-pointM[0]*pointN[1])-(pointM[1]-pointN[1])*(pointL[0]*pointK[1]-pointK[0]*pointL[1]))/((pointN[1]-pointM[1])*(pointL[0]-pointK[0])-(pointL[1]-pointK[1])*(pointN[0]-pointM[0])))
    return pointS

def draw_polylines_on_canvas(all_polylines, color):
    print("Drawing polylines on canvas")
    i = 0
    while (i < len(all_polylines)):
        j = 0
        while (j < len(all_polylines[i])-1):
            canvas.create_line(all_polylines[i][j][0],all_polylines[i][j][1],all_polylines[i][j+1][0],all_polylines[i][j+1][1], fill = color)
            j = j + 1
        canvas.create_line(all_polylines[i][j][0],all_polylines[i][j][1],all_polylines[i][0][0],all_polylines[i][0][1], fill = color)
        i = i + 1    


def set_X0_Y0(all_polylines):
    print("Setting xMin and yMin to zero")
    global xMin
    global yMin
    i = 0
    while (i < len(all_polylines)):
        j = 0
        while (j < len(all_polylines[i])):
            all_polylines[i][j][0] = all_polylines[i][j][0] - xMin
            all_polylines[i][j][1] = all_polylines[i][j][1] - yMin
            j = j + 1
        i = i + 1
    return all_polylines
            

def calculate_scaling():
    global xMin
    global xMax
    global yMin
    global yMax
    global xCanvas
    global yCanvas

    if (xCanvas / (xMax - xMin)) < (yCanvas / (yMax - yMin)):
        scaling = xCanvas / (xMax - xMin)
    else:
        scaling = yCanvas / (yMax - yMin)

    return scaling


def map_point_on_canvas(point):
    global xMin
    global yMin
    global yCanvas
    global scaling

    point[0] = int((point[0] - xMin) * scaling)
    point[1] = yCanvas - int((point[1] - yMin) * scaling)

    return point


def map_all_polylines_on_canvas(input_polylines):
    all_polylines = deepcopy(input_polylines)
    i = 0
    while (i < len(all_polylines)):
        j = 0
        while (j < len(all_polylines[i])):
            all_polylines[i][j] = map_point_on_canvas(all_polylines[i][j])
            j = j + 1         
        i = i + 1
    return all_polylines


def find_xMin(all_polylines):
    i = 0
    xMin = 9999999999
    while i < len(all_polylines):
        j = 0
        while j < len(all_polylines[i]):
            if all_polylines[i][j][0] < xMin:
                xMin = all_polylines[i][j][0]

            j = j + 1

        i = i + 1

    return xMin

        
def find_xMax(all_polylines):
    i = 0
    xMax = -9999999999
    while i < len(all_polylines):
        j = 0
        while j < len(all_polylines[i]):
            if all_polylines[i][j][0] > xMax:
                xMax = all_polylines[i][j][0]

            j = j + 1

        i = i + 1

    return xMax

        
def find_yMin(all_polylines):
    i = 0
    yMin = 9999999999
    while i < len(all_polylines):
        j = 0
        while j < len(all_polylines[i]):
            if all_polylines[i][j][1] < yMin:
                yMin = all_polylines[i][j][1]

            j = j + 1

        i = i + 1

    return yMin

        
def find_yMax(all_polylines):
    i = 0
    yMax = -9999999999
    while i < len(all_polylines):
        j = 0
        while j < len(all_polylines[i]):
            if all_polylines[i][j][1] > yMax:
                yMax = all_polylines[i][j][1]

            j = j + 1

        i = i + 1

    return yMax

        

def eliminate_parallel_lines(all_polylines):

    number_of_deleted_lines = 0
    min_length_diff = 0.001  #length of vector after subtraction from following line

    print("Eliminating lines, if the length of the vector, that results from the subtraction from the following line (both normalized) is shorter than "),
    print(min_length_diff),
    print("mm")

    i = 0
    while (i < len(all_polylines)):
        polyline = []
        j = 0
        while (j < len(all_polylines[i])):
            pointA = []
            pointB = []
            pointC = []

            pointA = all_polylines[i][j]

            if (j == len(all_polylines[i]) - 1):
                pointB = all_polylines[i][0]
            else:
                pointB = all_polylines[i][j+1]

            if (j < len(all_polylines[i]) - 2):
                pointC = all_polylines[i][j+2]
            else:
                if (j == len(all_polylines[i]) - 2):
                    pointC = all_polylines[i][0]
                else:
                    if (j == len(all_polylines[i]) - 1):
                        pointC = all_polylines[i][1]

            vectorA = []
            vectorA.append(pointB[1]-pointA[1])
            vectorA.append(pointB[0]-pointA[0])

            lengthA = line_length([[0,0], vectorA])

            vectorA[0] = vectorA[0] / lengthA
            vectorA[1] = vectorA[1] / lengthA

            vectorB = []
            vectorB.append(pointC[1]-pointB[1])
            vectorB.append(pointC[0]-pointB[0])

            lengthB = line_length([[0,0], vectorB])

            vectorB[0] = vectorB[0] / lengthB
            vectorB[1] = vectorB[1] / lengthB

            vector = []
            vector.append(vectorA[0] - vectorB[0])
            vector.append(vectorA[1] - vectorB[1])

            length = line_length([[0,0], vector])
            if length < min_length_diff:
                if (j == len(all_polylines[i]) - 1):
                    del(all_polylines[i][0])
                    number_of_deleted_lines = number_of_deleted_lines + 1
                    j = -1
                else:
                    del(all_polylines[i][j+1])
                    number_of_deleted_lines = number_of_deleted_lines + 1
                    j = -1

            j = j + 1

        i = i + 1
    print("Number of deleted lines: "),
    print(number_of_deleted_lines)



def eliminate_short_lines(all_polylines):
    i = 0
    line = []
    min_length = 0.01
    number_of_deleted_lines = 0


    print("Eliminating lines shorter "),
    print(min_length),
    print("mm")

    while (i < len(all_polylines)):
        Polylinie = []
        j = 0
        while (j < len(all_polylines[i])):
            if (j == len(all_polylines[i]) - 1):
                line = []
                line.append(all_polylines[i][j])
                line.append(all_polylines[i][0])
                if line_length(line) < min_length:
                    del(all_polylines[i][j])
                    number_of_deleted_lines = number_of_deleted_lines + 1
                    j = -1
            else:
                line = []
                line.append(all_polylines[i][j])
                line.append(all_polylines[i][j+1])
                if line_length(line) < min_length:
                    del(all_polylines[i][j+1])
                    number_of_deleted_lines = number_of_deleted_lines + 1
                    j = -1

            j = j + 1

        i = i + 1
    print("Number of deleted lines: "),
    print(number_of_deleted_lines)
                            

#Returns the length of a line
def line_length(line):
    return math.sqrt(pow(line[0][0]-line[1][0],2) + pow(line[0][1]-line[1][1],2))


# Returns the length of a file
def file_len(fname):
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
        f.close()
    return i + 1


# Works only with KiCAD 4.0.2 dxf-export. No general import function.
def dxf_import(fname):
    file_length = file_len(fname)

    print("Import "),
    print(filename_input)
    
    new_polyline = True
    line = ""
    polyline = []
    all_polylines = []

    linecount = 0

    file_input = open(fname)
    i = 0
    while ((i < file_length) and (line <> "ENTITIES")):
        i = i + 1
        line = file_input.readline()
        line = line.strip()   

    while i < file_length:
        i = i + 1
        line = file_input.readline()
        line = line.strip()

        if line == "LINE":
            linecount = linecount + 1
            pointA = []
            pointB = []

            line = file_input.readline()
            i = i + 1
            line = file_input.readline()
            i = i + 1
            line = file_input.readline()
            i = i + 1

            line = file_input.readline()
            i = i + 1
            line = line.strip()
            pointA.append(float(line))
            
            line = file_input.readline()
            i = i + 1

            line = file_input.readline()
            i = i + 1
            line = line.strip()
            pointA.append(float(line))
            
            line = file_input.readline()
            i = i + 1

            line = file_input.readline()
            i = i + 1
            line = line.strip()
            pointB.append(float(line))
            
            line = file_input.readline()
            i = i + 1

            line = file_input.readline()
            i = i + 1
            line = line.strip()
            pointB.append(float(line))
            

            if (new_polyline == False):
                if (pointA == polyline[len(polyline)-1]):
                    polyline.append(pointB)
                else:
                    new_polyline = True
            
            if new_polyline:
                if (polyline <> []):
                    all_polylines.append(polyline)
                polyline = []
                polyline.append(pointA)
                polyline.append(pointB)
                new_polyline = False

    if (new_polyline == False):
        all_polylines.append(polyline)

    file_input.close()

    print"Number of lines: " + str(linecount)
    print"Number of polylines: " + str(len(all_polylines))

    return(all_polylines)


filename_input = raw_input("dxf-File as input(default Test footprints-F.Cu.dxf)") or ("Test footprints-F.Cu.dxf")
filename_output = raw_input("ngc-File as output (default output.ngc)") or ("output.ngc")
file_length = 0
laser_diameter = float(raw_input("Assumed laserdiameter (default 0.2mm)") or ("0.2"))
distance_fillinglines = float(raw_input("Distance between filling lines (default 0.15mm)") or ("0.15"))
border = float(raw_input("Border around your layout (default 5mm)") or ("5"))
optimizing_cycles = int(raw_input("Number of optimizing cycles to reduce positioning (default 20)") or ("20"))
if (raw_input("Set xMin and yMin to zero (default Y)") or ("Y")) == "Y":
    set_xMin_yMin_to_zero = True
else:
    set_xMin_yMin_to_zero = False
  
geometry = dxf_import(filename_input)

xMin = find_xMin(geometry) - border
print("xMinimum = "),
print xMin

xMax = find_xMax(geometry) + border
print("xMaximum = "),
print xMax

yMin = find_yMin(geometry) - border
print("yMinimum = "),
print yMin

yMax = find_yMax(geometry) + border
print("yMaximum = "),
print yMax

scaling = calculate_scaling()
print("Scaling: "),
print(scaling)

if set_xMin_yMin_to_zero:
    set_X0_Y0(geometry)
    xMin = find_xMin(geometry) - border
    print("xMinimum = "),
    print xMin

    xMax = find_xMax(geometry) + border
    print("xMaximum = "),
    print xMax

    yMin = find_yMin(geometry) - border
    print("yMinimum = "),
    print yMin

    yMax = find_yMax(geometry) + border
    print("yMaximum = "),
    print yMax

    scaling = calculate_scaling()
    print("Scaling: "),
    print(scaling)


eliminate_short_lines(geometry)

eliminate_parallel_lines(geometry)

draw_polylines_on_canvas(map_all_polylines_on_canvas(geometry), "black")


geometriy_outlines = calculate_outline(geometry, laser_diameter)

optimizing_outline_arrangement(geometriy_outlines)

draw_polylines_on_canvas(map_all_polylines_on_canvas(geometriy_outlines), "red")


filling_pattern = calculate_filling_pattern(eliminate_vertical_lines(explode_to_lines(geometriy_outlines)))

draw_fillingpattern_on_canvas(map_fillingpattern_on_canvas(filling_pattern),"blue")

export_gcode(geometriy_outlines, filling_pattern)

mainloop()
